package backend.hdl.arch.rewrite.transpose

import backend.hdl.OrderedStreamTypeT
import backend.hdl.arch.rewrite.RulesHelper
import backend.hdl.arch.{ArchLambda, MapOrderedStream, SlideOrderedStream, TransposeNDOrderedStream}
import core.{ArithType, TypeChecker}
import core.rewrite.Rule

object SwapMapsRules {

  def exchangeABMapWithSize: Rule = Rule("exchangeABMapWithSize", {
    case MapOrderedStream(ArchLambda(p1, MapOrderedStream(ArchLambda(p2, body, _), in2, _), _), in1, _) if
      RulesHelper.containsParam(body, p1) && RulesHelper.containsParam(body, p2) &&
        // Avoid swapping expr generated by counters
        !RulesHelper.containsCounter(body) && !RulesHelper.isCounter(in1) && !RulesHelper.isCounter(in2) &&
        in1.t.asInstanceOf[OrderedStreamTypeT].bitWidth.ae.evalInt < in2.t.asInstanceOf[OrderedStreamTypeT].bitWidth.ae.evalInt =>
      val newMap = TypeChecker.check(MapOrderedStream(ArchLambda(p2, MapOrderedStream(ArchLambda(p1, body), in1)), in2))
      val dimLen = newMap.t.asInstanceOf[OrderedStreamTypeT].dimensions.length
      val dimOrder = RulesHelper.generateCountSeq(0,  dimLen-2) ++ Seq(ArithType(dimLen-1), ArithType(dimLen-2))
      TransposeNDOrderedStream(newMap, dimOrder)
  })

  def exchangeABMapSlide: Rule = Rule("exchangeABMapSlide", {
    case MapOrderedStream(ArchLambda(p1, MapOrderedStream(ArchLambda(p2, body, _), in2, _), _), SlideOrderedStream(in1, windowSize, _), _) if
      RulesHelper.containsParam(body, p1) && RulesHelper.containsParam(body, p2)  =>
      val newMap = TypeChecker.check(MapOrderedStream(ArchLambda(p2, MapOrderedStream(ArchLambda(p1, body), SlideOrderedStream(in1, windowSize))), in2))
      val dimLen = newMap.t.asInstanceOf[OrderedStreamTypeT].dimensions.length
      val dimOrder = RulesHelper.generateCountSeq(0,  dimLen-2) ++ Seq(ArithType(dimLen-1), ArithType(dimLen-2))
      TransposeNDOrderedStream(newMap, dimOrder)
  })
}
