package cGen.eqsat

import scala.collection.mutable

/**
  * Rewrites sub-graphs of e-graphs based on a context.
  */
trait ContextualSubgraphRewriter[ENode <: ENodeT, EClass <: EClassT, Context] {
  def eGraph: EGraphT[ENode, EClass]
  val builder: SubgraphBuilder[ENode, EClass]
  protected final val rewritten = new mutable.HashMap[(EClass, Context), builder.Promise]()

  protected def childContext(node: ENode, child: EClass, context: Context): Context
  protected def rewriteAndAdd(node: ENode, context: Context): builder.Promise
  protected def initialContext(eClass: EClass): Context

  protected final def rewriteAndAddDefault(node: ENode, context: Context): builder.Promise = {
    val children = node.args.map(arg =>
      rewrite(
        arg.asInstanceOf[EClass],
        childContext(node, arg.asInstanceOf[EClass], context)))
    builder.add(node.expr, children)
  }

  protected def rewrite(eClass: EClass, context: Context): builder.Promise = {
    val leader = eGraph.find(eClass)
    rewritten.get((leader, context)) match {
      case Some(precomputed) => return precomputed
      case None =>
    }

    val newLeader = builder.addEmpty(leader.t)
    rewritten((leader, context)) = newLeader

    for (node <- leader.nodes) {
      builder.merge(newLeader, rewriteAndAdd(node.asInstanceOf[ENode], context))
    }

    newLeader
  }

  def rewrite(eClass: EClass): builder.Promise = {
    rewrite(eClass, initialContext(eClass))
  }
}
