package cGen.eqsat

import core.Expr

/**
  * An extractor produces concrete expressions from e-classes.
  */
trait Extractor {
  /**
    * Attempts to recover a concrete expression from an e-class.
    * @param eClass The e-class to extract an expression from.
    * @return An expression equivalent to `eClass` or `None`.
    */
  def apply(eClass: EClassT): Option[Expr]
}

object Extractor {
  /**
    * The default extractor, which produces smallest expressions.
    */
  object DefaultSmallest extends Extractor {
    override def apply(eClass: EClassT): Option[Expr] = {
      if (eClass.hasAnalysisResult(ExtractionAnalysis.smallestExpressionExtractor)) {
        Some(eClass.analysisResult(ExtractionAnalysis.smallestExpressionExtractor).get._1)
      } else {
        DirectExtractor(CostFunction.Size, _ => true, eClass.reachable).apply(eClass)
      }
    }
  }
}
