package cGen.eqsat

import scala.collection.mutable.ArrayBuffer

/**
  * An observer that counts the number of nodes in the e-graph at every saturation step.
  */
class NodesOverTimeObserver extends StepObserver {
  private val nodeCounts = new ArrayBuffer[Int]()

  /**
    * The number of nodes over time, as measured by this observer.
    * @return The number of nodes over time.
    */
  def nodesOverTime: Seq[Int] = nodeCounts

  /**
    * Notes that the graph is about to perform a step in the Equality Saturation algorithm.
    * @param graph The graph that is about to take a step
    */
  override def onStepOrFinish[ENode <: ENodeT, EClass <: EClassT](graph: EGraphT[ENode, EClass]): Unit = {
    nodeCounts.append(graph.classes.map(_.nodes.length).sum)
  }
}
