package cGen.eqsat

import core.{Expr, LambdaT}

trait ParameterUseRewriter[ENode <: ENodeT, EClass <: EClassT, Value] extends ContextualSubgraphRewriter[ENode, EClass, List[(Value, Int)]] {
  protected def lambdaValue(lambda: LambdaT): Value

  override def childContext(node: ENode, child: EClass, context: List[(Value, Int)]): List[(Value, Int)] = {
    // Figure out which parameters appear in the subgraph we're about to rewrite.
    val uses = child.analysisResult(ParameterUseAnalysis)

    val newContext = node.expr match {
      case lambda: LambdaT => (lambdaValue(lambda), 0) :: context.map(t => (t._1, t._2 + 1))
      case DeBruijnShift(_) => context.tail.map(t => (t._1, t._2 - 1))
      case _ => context
    }

    // Truncate the context, dropping all value assignments to parameters that we won't use.
    val highestIndex = (0 :: uses.map(_.index + 1).toList).max
    newContext.take(highestIndex)
  }
}
