package cGen.eqsat

/**
  * A set of samples that are being evaluated. The set itself decides when it has received enough samples, as well as
  * which sample is best.
  */
trait SampleEvaluationSet[A] {
  /**
    * Adds a sample to this evaluation set.
    * @param sample The sample to return.
    */
  def add(sample: A): Unit

  /**
    * Alerts the sample evaluation set that sampling failed.
    */
  def onSampleFailure(): Unit

  /**
    * Asks the sample evaluation set if it has received enough samples.
    * @return `true` if the set wants no further samples; `false` if it wants more samples.
    */
  def isSaturated: Boolean

  /**
    * The best sample in the set, according to some metric.
    * @return A sample in this set or `None`.
    */
  def best: Option[A]
}
