package cGen.idioms

import algo.AlgoFunType
import cGen.CGenExpr
import core._

case class IntrinsicCallExpr(innerIR: Expr, name: String, ins: Seq[Expr], intvs: Seq[TypeVarT], outt: Type) extends CGenExpr {
  override def types: Seq[Type] = Seq()

  override def cgenBuild(newInnerIR: Expr, newTypes: Seq[Type]): IntrinsicCallExpr =
    IntrinsicCallExpr(newInnerIR, name, ins, intvs, outt)

  override protected def argsBuild(args: Seq[Expr]): CGenExpr =
    IntrinsicCall(name, args, intvs, outt)
}

object IntrinsicCall { // the input intvs/outt/ins/ should be reversed
  def apply(name: String, ins: Seq[Expr], intvs: Seq[TypeVarT], outt: Type): IntrinsicCallExpr = {
    IntrinsicCallExpr({
      FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(intvs, AlgoFunType(intvs, outt))), ins.map(_.t)), ins)
    }, name, ins, intvs, outt)
  }

  def unapply(expr: IntrinsicCallExpr): Some[(String, Seq[Expr], Seq[TypeVarT], Type)] = {
    Some(expr.name, expr.ins.indices.map(expr.args(_)), expr.intvs, expr.outt)
  }
}