package backend.hdl

import algo.{Input, Join, Repeat}
import algo.util.Matrix
import backend.hdl.arch.mem.{MemFunctionsCompiler, MemoryImage}
import backend.hdl.arch.rewrite.{CleanupRules, InputBufferingRules}
import core.Counter
import core.compile.CompilerPhase
import core.rewrite.{RewriteAll, RewriteStep, RewriteTargeted}
import org.junit.Test

class FastWriteTest {
  def repeatInner(A: Seq[Int], repetitions: Int): Seq[Int] = {
    val result: Array[Int] = Array.ofDim[Int](A.length * repetitions)
    for (i <- A.indices) {
      for (j <- 0 until repetitions) {
        result(i * repetitions + j) = A(i)
      }
    }
    result.toSeq
  }

  @Test
  def fastWriteTest(): Unit = {
    val cachelines = 128
    val repeatTimes = 16
    val data = Matrix.fromCounter(1, cachelines, 1, Some("input")).data.head
    val res = repeatInner(data, repeatTimes)
    Counter.resetAll()
    val alg = Join(algo.Map(Repeat.asFunction(Seq(None), Seq(repeatTimes)), Input("data", algo.IntType(512), cachelines)))
    val proj = HDLProject("repeatV2", alg, CompilerPhase.first(), Seq(
      (MemFunctionsCompiler.phaseAfter, RewriteStep(RewriteAll(), Seq(InputBufferingRules.limitParallelReadRequests(64)))),
      (MemFunctionsCompiler.phaseAfter, RewriteStep(RewriteAll(), CleanupRules.get())),
      (MemFunctionsCompiler.phaseAfter, RewriteStep(RewriteTargeted(0), Seq(InputBufferingRules.doubleBufferRead)))
    ))

    proj.simulateWithHostRam(
      Predef.Map(
        "data" -> data,
        MemoryImage.RESULT_VAR_NAME -> res
      )
    ).print()
  }
}
