package backend.hdl.arch

import algo.util.Matrix
import algo.{Input, MMMul}
import backend.hdl.HDLProject
import backend.hdl.arch.device.DeviceSpecificCompiler
import backend.hdl.arch.mem.{MemFunctionsCompiler, MemoryImage}
import backend.hdl.arch.rewrite._
import backend.hdl.arch.rewrite.tiling.{ABTilingRules, FoldTilingRules}
import backend.hdl.arch.tiling.{ParallelizationCompiler, TranspositionCompiler}
import core.compile.CompilerPhase
import core.rewrite.{RewriteAll, RewriteStep, RewriteTargeted}
import core.Counter
import org.junit.Test

class MxMTilingTest {

  def tiledMatrices(rows: Int, cols: Int, tileSize: Int): Map[String, Seq[Seq[Int]]] = {
    val matA = Matrix.fromRandom(rows, cols, 9, Some("A"))
    val matB = Matrix.fromRandom(rows, cols, 9, Some("B"))
    val matC = matA.tiledMul(matB, tileSize, false)
    Predef.Map(
      "matrix1" -> matA.data,
      "matrix2" -> matB.data,
      MemoryImage.RESULT_VAR_NAME -> matC.data
    )
  }

  @Test
  def testMxMTiling(): Unit = {
    Counter.resetAll()
    val outerTile = 128 // 512
    val innerTile = 128 // 2048
    val imageSize = 256 // 4096
    val inputElementBitWidth = 8
    val alg = MMMul(Input("matrix1", imageSize, imageSize, algo.IntType(inputElementBitWidth)), Input("matrix2", imageSize, imageSize, algo.IntType(inputElementBitWidth)))

    HDLProject("mxm_tiling", alg, CompilerPhase.first(), Seq(
      (ArchCompiler.phaseBefore, RewriteStep(RewriteTargeted(0), Seq(ABTilingRules.tilingInnerMap(0, outerTile)))),
      (ArchCompiler.phaseBefore, RewriteStep(RewriteTargeted(0), Seq(ABTilingRules.exchangeABMapNoCondition(1)))),
      (ArchCompiler.phaseBefore, RewriteStep(RewriteTargeted(0), Seq(ABTilingRules.tilingInnerMap(0, outerTile)))),
      (ArchCompiler.phaseBefore, RewriteStep(RewriteTargeted(0), Seq(ABTilingRules.exchangeABMapNoCondition(1)))),
      (ArchCompiler.phaseBefore, RewriteStep(RewriteTargeted(0), Seq(FoldTilingRules.tilingInsideFold(innerTile)))),

      (TranspositionCompiler.phaseAfter, RewriteStep(RewriteAll(), ParallelizeDotProductRules.get(Some(innerTile)))),
      (ParallelizationCompiler.phaseAfter, RewriteStep(RewriteAll(), IntermediateBufferingRules.get())),

      (MemFunctionsCompiler.phaseAfter, RewriteStep(RewriteAll(), InputBufferingRules.readDoubleBuffering)),
      (DeviceSpecificCompiler.phaseAfter, RewriteStep(RewriteAll(), FixTimingRules.get())),

    )).simulateWithHostRam(tiledMatrices(imageSize, imageSize, outerTile)).print()
  }
}
