package cGen

import cGen.Lib._
import cGen.TupleStructName.COMPLEX
import cGen.pass.TuningPass.DetermineViewTypePass
import cGen.primitive.SelectS
import core.{Expr, ParamDef, ParamUse, TypeChecker}
import org.junit.Test

import scala.Console.println

class FFTTest {
  implicit val fftT = () => DoubleType()

  def kronIdR(m: Int, f: Expr => Expr): Expr => Expr = v => {
    var tmp: Expr = Split(v, m)
    tmp = Transpose(tmp)
    tmp = f(tmp)
    tmp = Transpose(tmp)
    Join(tmp)
  }

  def mapMultPiT(r: Int): Expr => Expr = v => {
    val t = TypeChecker.check(v).t.asInstanceOf[ArrayTypeT].et
    val p = ParamDef(t)
    Map(CLambda(p, {
      var tmp: Expr = Split(ParamUse(p), r)
      tmp = Transpose(tmp)
      Join(tmp)
    }), v)
  }

  def stPerm(r: Int, L: Int, n: Int): Expr => Expr = {
    v => kronIdR(n / L, mapMultPiT(r))(v)
  }

  @Test
  def stPermTest(): Unit = {
    val r = 2
    val L = 4
    val n = 8
    val v = ParamDef(ArrayType(TupleType(COMPLEX, fftT(), fftT()), n))
    var tree: Expr = CLambda(v, Materialize(stPerm(r, L, n)(ParamUse(v))))
    tree = DetermineViewTypePass(tree).asInstanceOf[Expr]
    tree = DPSPass(tree)
//    CCodeGen(tree).genTest("st_perm")
    val code = CCodeGen(tree, "st_perm").genTest()
    println(code)
  }

  def mapMultB(r: Int, L: Int): Expr => Expr = v => { // not finished
    val t = TypeChecker.check(v).t.asInstanceOf[ArrayTypeT].et
    val p = ParamDef(t)
    //        val prog = new Program()
    //        val tmp = prog << Materialize(Zip(Tuple(
    //          Transpose(Split(ParamUse(p), L / r)),
    //          RepOfB(r, L))))
    //    prog build Map(CLambda(p, {
    Map(CLambda(p, {
      var tmp: Expr = Zip(Tuple(
        Transpose(Split(ParamUse(p), L / r, SrcViewType()), SrcViewType()),
        RepOfB(r, L), SrcViewType()), SrcViewType())
      val w = ParamDef(TypeChecker.check(tmp).t.asInstanceOf[ArrayTypeT].et)
      tmp = Map(CLambda(w, {
        complexMatVecMul(SelectS(ParamUse(w), 1), SelectS(ParamUse(w), 0))
      }), tmp, SrcViewType())
      tmp = Join(Transpose(tmp, SrcViewType()), SrcViewType())
      //      println(TypeChecker.check(tmp).t)
      TypeChecker.check(tmp)
      tmp
    }), v, SrcViewType())
  }

  def RepOfB(r: Int, L: Int): Expr = {
    val j = ParamDef(IntType(32))
    val k = ParamDef(IntType(32))
    val l = ParamDef(IntType(32))
    val numMats = Constant(r, fftT())
    Build(CLambda(j,
      Build(CLambda(k,
        Build(CLambda(l, // not sure about the input omega
          omega(
            mul(add(mul(TypeConv(ParamUse(k), fftT()), numMats), TypeConv(ParamUse(j), fftT())), TypeConv(ParamUse(l), fftT())),
            Constant(r * r, fftT()))
        ), r, SrcViewType())
      ), r, SrcViewType())
    ), L / r, SrcViewType())
  }

  def omega(k: Expr, n: Expr): Expr = {
    val a = Constant(-2 * math.Pi, fftT())
    ComplexTuple(cos(div(mul(a, k), n)), cos(div(mul(a, k), n)), Some(DestViewType()))
  }

  def stCombine(r: Int, L: Int, n: Int): Expr => Expr = {
    val m = n / L
    v => {
      val prog = new Program()
      var tmp: Expr = Split(v, m)
      tmp = Transpose(tmp)
      tmp = mapMultB(r, L)(tmp)
      tmp = prog << Materialize(tmp)
      tmp = Transpose(tmp)
      tmp = Join(tmp)
      tmp = Materialize(tmp)
      prog build tmp
    }
    //      kronIdR(n / L, mapMultB(r, L))
  }

  @Test
  def stCombineTest(): Unit = {
    val r = 2
    val L = 4 // L = 2 or 2 * 2 or 2 * 2 * 2
    val n = 8
    val v = ParamDef(ArrayType(TupleType(COMPLEX, fftT(), fftT()), n))
    var tree: Expr = CLambda(v, stCombine(r, L, n)(ParamUse(v)))
    tree = DetermineViewTypePass(tree).asInstanceOf[Expr]
    tree = TypeChecker.check(tree)
    tree = DPSPass.apply(tree)
    val code = CCodeGen(tree, "st_combine").genTest()
    println(code)
  }

  @Test
  def complexVecMulTest(): Unit = {
    val r = 2
    val L = 4
    val n = 8
    val v0 = ParamDef(ArrayType(TupleType(COMPLEX, fftT(), fftT()), n))
    val v1 = ParamDef(ArrayType(TupleType(COMPLEX, fftT(), fftT()), n))
    val mat = ParamDef(ArrayType(ArrayType(TupleType(COMPLEX, fftT(), fftT()), n), n))
    var tree: Expr = CLambda(v0, CLambda(v1, complexVecMul(ParamUse(v0), ParamUse(v1))))
    tree = DetermineViewTypePass(tree).asInstanceOf[Expr]
    tree = DPSPass.apply(tree)
    val k = CCodeGen(tree, "complex_vec_mul")
    println(k)

//    tree = DPSPass(tree)

  }

  @Test
  def MapMultBTest(): Unit = { // works only when switch to the old typechecker
    val r = 2
    val L = 4
    val n = 8
    val v0 = ParamDef(ArrayType(TupleType(COMPLEX, fftT(), fftT()), n))
    val v1 = ParamDef(ArrayType(TupleType(COMPLEX, fftT(), fftT()), n))
    val mat2 = ParamDef(ArrayType(ArrayType(TupleType(COMPLEX, fftT(), fftT()), n / 2), n / 2))
    var tree: Expr = CLambda(mat2, mapMultB(2, 4)(ParamUse(mat2)))
    tree = DetermineViewTypePass(tree).asInstanceOf[Expr]
    tree = DPSPass(tree)
    println(CCodeGen(tree))
  }
}
