package cGen

import cGen.Lib.{AngleAxisToRotationMatrix, EulerAnglesToRotationMatrix, GMM, MakeRelative, ProjectAndSqNorm}
import cGen.idioms.{IdiomRecognition, LowerIntrinsics}
import core.{Expr, ParamDef, ParamUse, TypeChecker}
import org.junit.{Ignore, Test}

@Ignore
class IdiomRecognitionTest {
  @Test
  def recognizeDotProduct(): Unit = {
    implicit val t: () => CScalarTypeT = DoubleType
    val m1 = ParamDef(ArrayType(DoubleType(), 100))
    val m2 = ParamDef(ArrayType(DoubleType(), 100))
    var tree: Expr = CLambda(m1, CLambda(m2, Lib.DotProd(ParamUse(m1), ParamUse(m2))))
    tree = IdiomRecognition(tree)
    tree = LowerIntrinsics(tree)
    tree = DPSPass(tree)
    println(CCodeGen(tree).ccode())
  }

  @Test
  def recognizeMatrixMul(): Unit = {
    implicit val t: () => CScalarTypeT = DoubleType
    val m1 = ParamDef(ArrayType(ArrayType(t(), 3), 3))
    val m2 = ParamDef(ArrayType(ArrayType(t(), 4), 3))
    var tree: Expr = CLambda(m1, CLambda(m2, Lib.MatrixMul(ParamUse(m1), ParamUse(m2))))
    tree = idioms.IdiomRecognition(tree)
    tree = idioms.LowerIntrinsics(tree)
    tree = DPSPass(tree)
    val ccode = CCodeGen(tree)
    println(ccode)
  }

  @Test
  def simplifyPretransposedMatrixMul(): Unit = {
    implicit val t: () => CScalarTypeT = DoubleType
    val m1 = ParamDef(ArrayType(ArrayType(t(), 3), 3))
    val m2 = ParamDef(ArrayType(ArrayType(t(), 3), 3))
    val impl = Lib.MatrixMul(ParamUse(m1), Lib.transpose(ParamUse(m2)))
    var tree: Expr = CLambda(m1, CLambda(m2, impl))
    tree = idioms.IdiomRecognition(tree)
    tree = idioms.LowerIntrinsics(tree)
    tree = DPSPass(tree)
    val ccode = CCodeGen(tree)
    println(ccode)
  }

  @Test
  def projectAndSqNormTest(): Unit = {
    implicit val t: () => DoubleType = () => DoubleType()
    val cam = ParamDef(ArrayType(DoubleType(), 11))
    val x = ParamDef(ArrayType(DoubleType(), 3))
    var tree: Expr = CLambda(cam, CLambda(x, ProjectAndSqNorm(ParamUse(cam), ParamUse(x))))
    tree = idioms.IdiomRecognition(tree)
    tree = idioms.LowerIntrinsics(tree)
    tree = DPSPass(tree)
    val ccode = CCodeGen(tree)
    println(ccode)
  }

  @Test
  def GMMTest(): Unit = {
    implicit val t = () => DoubleType()
    val n = 100
    val d = 3
    val K = 5
    val td = ((d) * ((d) + (1))) / (2)
    val x = ParamDef(ArrayType(ArrayType(t(), d), n))
    val alphas = ParamDef(ArrayType(t(), K))
    val means = ParamDef(ArrayType(ArrayType(t(), d), K))
    val qs = ParamDef(ArrayType(ArrayType(t(), d), K))
    val ls = ParamDef(ArrayType(ArrayType(t(), td), K))
    val wishartGamma = ParamDef(t())
    val wishartM = ParamDef(t())
    var tree: Expr = CLambda(x, CLambda(alphas, CLambda(means, CLambda(qs, CLambda(ls, CLambda(wishartGamma, CLambda(wishartM,
      GMM(n, d, K, ParamUse(x), ParamUse(alphas), ParamUse(means), ParamUse(qs), ParamUse(ls), ParamUse(wishartGamma), ParamUse(wishartM)))))))))
    tree = idioms.IdiomRecognition(tree)
    tree = idioms.LowerIntrinsics(tree)
    tree = DPSPass(tree)
    val ccode = CCodeGen(tree)
    println(ccode)
  }

  @Test
  def angleAxisToRotationMatrixTest(): Unit = {
    val angleAxis = ParamDef(ArrayType(DoubleType(), 3))
    var tree: Expr = CLambda(angleAxis, AngleAxisToRotationMatrix(ParamUse(angleAxis)))
    tree = idioms.IdiomRecognition(tree)
    tree = idioms.LowerIntrinsics(tree)
    tree = DPSPass(tree)
    val ccode = CCodeGen(tree)
    println(ccode)
  }

  @Test
  def eulerAnglesToRotationMatrixTest(): Unit = {
    val eulerAngles = ParamDef(ArrayType(DoubleType(), 3))
    var tree: Expr = CLambda(eulerAngles, EulerAnglesToRotationMatrix(ParamUse(eulerAngles)))
    tree = idioms.IdiomRecognition(tree)
    tree = idioms.LowerIntrinsics(tree)
    tree = DPSPass(tree)
    val ccode = CCodeGen(tree)
//    ccode.genTest("angle_axis_rotation_matrix")
    println(ccode)
  }

  @Test
  def makeRelativeTest(): Unit = {
    val poseParam = ParamDef(ArrayType(DoubleType(), 3))
    val baseRelative = ParamDef(ArrayType(ArrayType(DoubleType(), 4), 4))
    var tree: Expr = CLambda(poseParam, CLambda(baseRelative, MakeRelative(ParamUse(poseParam), ParamUse(baseRelative))))
    tree = idioms.IdiomRecognition(tree)
    tree = idioms.LowerIntrinsics(tree)
    tree = TypeChecker.check(tree)
    tree = TypeChecker.check(tree)
//    tree = ExternalCallCopyInsertion(tree)
    tree = DPSPass(tree)
    val ccode = CCodeGen(tree)
//    ccode.genTest("make_relative")
    println(ccode)
  }
}
