package cGen

import cGen.Lib.{Program, VectorSum, add, conv2, div, five, mul, nine, one, seven, slideS2D, sub, transpose, two, zero}
import cGen.primitive.{IdxS, RollingSlide}
import core.util.IRDotGraph
import core.{ArithType, ArithTypeT, Expr, ParamDef, ParamUse, Type, TypeChecker}
import org.junit.Test

class OpenCLGenTest {
  @Test
  def jacobi1d(): Unit = {
    implicit val t = () => FloatType()
    val len = 4096
    val in = ParamDef(ArrayType(t(), len))
    val w = ParamDef(ArrayType(t(), 3))
    var tree: Expr = CLambda(in, MapGlb(
      CLambda(w, mul(VectorSum(ParamUse(w)), Constant(0.33333, t()))), Slide(ParamUse(in), 3)
    ))
    tree = DPSPass(tree)
    val ccode = CCodeGen(tree, "cl_kernel", platform = OpenCL64).genCode("jacobi1D_opencl")
    print(ccode)
  }

  @Test
  def jacobi2d(): Unit = {
    implicit val t = () => FloatType()
    val len = 4096
    val in = ParamDef(ArrayType(ArrayType(t(), len), len))
    val w = ParamDef(ArrayType(ArrayType(t(), 3), 3))
    var tree: Expr = CLambda(in,
      MapGlb2(conv2, slideS2D(ParamUse(in), 3))
    )
    tree = DPSPass(tree)
    val ccode = CCodeGen(tree, "cl_kernel", platform = OpenCL64).genCode("jacobi2D_opencl")
    print(ccode)
  }

  def MapGlb2(f: Expr, in: Expr) = {
    val p0 = ParamDef(TypeChecker.check(in).t.asInstanceOf[ArrayTypeT].et)
    val p1 = ParamDef(TypeChecker.check(in).t.asInstanceOf[ArrayTypeT].et.asInstanceOf[ArrayTypeT].et)
    MapGlb(CLambda(p0, MapGlb(CLambda(p1, CFC(f, ParamUse(p1))), ParamUse(p0))), in)
  }

  def MapGlb3(f: Expr, in: Expr) = {
    val p0 = ParamDef(TypeChecker.check(in).t.asInstanceOf[ArrayTypeT].et)
    val p1 = ParamDef(TypeChecker.check(in).t.asInstanceOf[ArrayTypeT].et.asInstanceOf[ArrayTypeT].et)
    val p2 = ParamDef(TypeChecker.check(in).t.asInstanceOf[ArrayTypeT].et.asInstanceOf[ArrayTypeT].et.asInstanceOf[ArrayTypeT].et)
    MapGlb(CLambda(p0, MapGlb(CLambda(p1, MapGlb(CLambda(p2, CFC(f, ParamUse(p2))), ParamUse(p1))), ParamUse(p0))), in)
  }

  def Map2(f: Expr, in: Expr, vt: ViewTypeT = EagerType()) = {
    val p0 = ParamDef(TypeChecker.check(in).t.asInstanceOf[ArrayTypeT].et)
    val p1 = ParamDef(TypeChecker.check(in).t.asInstanceOf[ArrayTypeT].et.asInstanceOf[ArrayTypeT].et)
    Map(CLambda(p0, Map(CLambda(p1, {
      CFC(f, ParamUse(p1))
    }), ParamUse(p0), vt)), in, vt)
  }

  def conv3d7pt(implicit datat: () => Type): CLambda = {
    implicit val t = () => IntType(32)
    val wd = ParamDef(ArrayType(ArrayType(ArrayType(datat().asInstanceOf[CGenDataTypeT], 3), 3), 3))
    CLambda(wd, {
      val prog = new Program()
      val sum = Seq(
        IdxS(IdxS(IdxS(ParamUse(wd), one), one), zero),
        IdxS(IdxS(IdxS(ParamUse(wd), zero), one), one),
        IdxS(IdxS(IdxS(ParamUse(wd), one), zero), one),
        IdxS(IdxS(IdxS(ParamUse(wd), one), one), one),
        IdxS(IdxS(IdxS(ParamUse(wd), one), two), one),
        IdxS(IdxS(IdxS(ParamUse(wd), two), one), one),
        IdxS(IdxS(IdxS(ParamUse(wd), one), one), two),
      ) map (e => prog.add(e)) reduce ((x: Expr, y: Expr) => add(x, y))
      prog.build(div(sum, seven(() => datat())))
    })
  }

  def conv3d15pt(implicit datat: () => Type): CLambda = {
    implicit val t = () => IntType(32)
    val wd = ParamDef(ArrayType(ArrayType(ArrayType(datat().asInstanceOf[CGenDataTypeT], 3), 3), 3))
    CLambda(wd, {
      val prog = new Program()
      val sum = Seq(
        IdxS(IdxS(IdxS(ParamUse(wd), one), zero), zero),
        IdxS(IdxS(IdxS(ParamUse(wd), one), zero), one),
        IdxS(IdxS(IdxS(ParamUse(wd), one), zero), two),

        IdxS(IdxS(IdxS(ParamUse(wd), one), one), zero),
        IdxS(IdxS(IdxS(ParamUse(wd), one), one), one),
        IdxS(IdxS(IdxS(ParamUse(wd), one), one), two),

        IdxS(IdxS(IdxS(ParamUse(wd), one), two), zero),
        IdxS(IdxS(IdxS(ParamUse(wd), one), two), one),
        IdxS(IdxS(IdxS(ParamUse(wd), one), two), two),

        IdxS(IdxS(IdxS(ParamUse(wd), zero), one), zero),
        IdxS(IdxS(IdxS(ParamUse(wd), zero), one), one),
        IdxS(IdxS(IdxS(ParamUse(wd), zero), one), two),

        IdxS(IdxS(IdxS(ParamUse(wd), two), one), zero),
        IdxS(IdxS(IdxS(ParamUse(wd), two), one), one),
        IdxS(IdxS(IdxS(ParamUse(wd), two), one), two),
      ) map (e => prog.add(e)) reduce ((x: Expr, y: Expr) => add(x, y))
      prog.build(div(sum, seven(() => datat())))
    })
  }

  def conv3d27pt(implicit datat: () => Type): CLambda = {
    implicit val t = () => IntType(32)
    val wd = ParamDef(ArrayType(ArrayType(ArrayType(datat().asInstanceOf[CGenDataTypeT], 3), 3), 3))
    CLambda(wd, {
      val prog = new Program()
      val sum = Seq(
        IdxS(IdxS(IdxS(ParamUse(wd), zero), zero), zero),
        IdxS(IdxS(IdxS(ParamUse(wd), zero), zero), one),
        IdxS(IdxS(IdxS(ParamUse(wd), zero), zero), two),

        IdxS(IdxS(IdxS(ParamUse(wd), zero), one), zero),
        IdxS(IdxS(IdxS(ParamUse(wd), zero), one), one),
        IdxS(IdxS(IdxS(ParamUse(wd), zero), one), two),

        IdxS(IdxS(IdxS(ParamUse(wd), zero), two), zero),
        IdxS(IdxS(IdxS(ParamUse(wd), zero), two), one),
        IdxS(IdxS(IdxS(ParamUse(wd), zero), two), two),

        IdxS(IdxS(IdxS(ParamUse(wd), one), zero), zero),
        IdxS(IdxS(IdxS(ParamUse(wd), one), zero), one),
        IdxS(IdxS(IdxS(ParamUse(wd), one), zero), two),

        IdxS(IdxS(IdxS(ParamUse(wd), one), one), zero),
        IdxS(IdxS(IdxS(ParamUse(wd), one), one), one),
        IdxS(IdxS(IdxS(ParamUse(wd), one), one), two),

        IdxS(IdxS(IdxS(ParamUse(wd), one), two), zero),
        IdxS(IdxS(IdxS(ParamUse(wd), one), two), one),
        IdxS(IdxS(IdxS(ParamUse(wd), one), two), two),

        IdxS(IdxS(IdxS(ParamUse(wd), two), zero), zero),
        IdxS(IdxS(IdxS(ParamUse(wd), two), zero), one),
        IdxS(IdxS(IdxS(ParamUse(wd), two), zero), two),

        IdxS(IdxS(IdxS(ParamUse(wd), two), one), zero),
        IdxS(IdxS(IdxS(ParamUse(wd), two), one), one),
        IdxS(IdxS(IdxS(ParamUse(wd), two), one), two),

        IdxS(IdxS(IdxS(ParamUse(wd), two), two), zero),
        IdxS(IdxS(IdxS(ParamUse(wd), two), two), one),
        IdxS(IdxS(IdxS(ParamUse(wd), two), two), two),
      ) map (e => prog.add(e)) reduce ((x: Expr, y: Expr) => add(x, y))
      prog.build(div(sum, seven(() => datat())))
    })
  }

  def transpose2(in: Expr): Expr = {
    val p = ParamDef(TypeChecker.check(in).t.asInstanceOf[ArrayTypeT].et)
    transpose(Map(CLambda(p, transpose(ParamUse(p))), in, SrcViewType()))
  }

  def mapTranspose(in: Expr): Expr = {
    val p = ParamDef(TypeChecker.check(in).t.asInstanceOf[ArrayTypeT].et)
    Map(CLambda(p, Transpose(ParamUse(p), SrcViewType())), in, SrcViewType())
  }

  def slideS3D(in: Expr, size: ArithTypeT): Expr = {
    val slide2 = slideS2D(in, size)
    val p = ParamDef(TypeChecker.check(slide2).t.asInstanceOf[ArrayTypeT].et.asInstanceOf[ArrayTypeT].et)
    Map2(CLambda(p, Slide(transpose2(ParamUse(p)), 3)), slide2, SrcViewType())
  }

  def stencil3d(conv3: Expr) = {
    implicit val t = () => FloatType()
    val len = 128
    val in = ParamDef(ArrayType(ArrayType(ArrayType(t(), len), len), len))
    CLambda(in, MapGlb3(conv3, slideS3D(ParamUse(in), 3)))
  }

  def rollingStencil3d(conv3: Expr): CLambda = {
    implicit val t = () => FloatType()
    val len = 128
    val in = ParamDef(ArrayType(ArrayType(ArrayType(t(), len), len), len))
    val w = ParamDef(ArrayType(ArrayType(ArrayType(t(), len), 3), 3))
    CLambda(in, {
      MapGlb2(CLambda(w, {
        val in = RollingSlide(transpose2(ParamUse(w)), 3, SrcViewType())
        val p = ParamDef(TypeChecker.check(in).t.asInstanceOf[ArrayTypeT].et)
        Map(CLambda(p, {
          val prog = new Program()
          val x = prog << ParamUse(p) // Common subexpression elimination
          prog build CFC(conv3, x)
        }), in, EagerType())
      }), slideS2D(ParamUse(in), 3))
    })
  }

  @Test
  def jacobi3d7pt(): Unit = {
    implicit val t = () => FloatType()
    val tree: Expr = DPSPass(stencil3d(conv3d7pt))
    val ccode = CCodeGen(tree, "cl_kernel", platform = OpenCL64).genCode("jacobi3d7pt_opencl")
    println(ccode)
  }

  @Test
  def jacobi3d15pt(): Unit = {
    implicit val t = () => FloatType()
    val tree: Expr = DPSPass(stencil3d(conv3d15pt))
    val ccode = CCodeGen(tree, "cl_kernel", platform = OpenCL64).genCode("jacobi3d15pt_opencl")
    println(ccode)
  }

  @Test
  def jacobi3d27pt(): Unit = {
    implicit val t = () => FloatType()
    val tree: Expr = DPSPass(stencil3d(conv3d27pt))
    val ccode = CCodeGen(tree, "cl_kernel", platform = OpenCL64).genCode("jacobi3d27pt_opencl")
    println(ccode)
  }

  @Test
  def rollingJacobi3d7pt(): Unit = {
    implicit val t = () => FloatType()
    val tree = DPSPass(rollingStencil3d(conv3d7pt))
    println(TypeChecker.check(tree).t)
    val ccode = CCodeGen(tree, "cl_kernel", platform = OpenCL64).genCode("rolling_jacobi3d7pt_opencl")
    println(ccode)
  }

  @Test
  def rollingJacobi3d15pt(): Unit = {
    implicit val t = () => FloatType()
    val tree = DPSPass(rollingStencil3d(conv3d15pt))
    println(TypeChecker.check(tree).t)
    val ccode = CCodeGen(tree, "cl_kernel", platform = OpenCL64).genCode("rolling_jacobi3d15pt_opencl")
    println(ccode)
  }

  @Test
  def rollingJacobi3d27pt(): Unit = {
    implicit val t = () => FloatType()
    val tree = DPSPass(rollingStencil3d(conv3d27pt))
    println(TypeChecker.check(tree).t)
    val ccode = CCodeGen(tree, "cl_kernel", platform = OpenCL64).genCode("rolling_jacobi3d27pt_opencl")
    println(ccode)
  }
}
