package cGen

import cGen.Lib.{add, sqrt}
import cGen.primitive.{Id, IdxS}
import core.{Expr, ParamDef, ParamUse, TypeChecker}
import org.junit.Test


class ViewTypeTest {

  val letp = ParamDef(IntType(32))
  // letEager is a let expression returns a call to eager primitive
  val letEager = CGenLet(ParamDef(IntType(32)), Id(ParamUse(letp)), IdxS(ParamUse(letp), Constant(1, IntType(32))))
  // letLazy is a let expression returns a call to lazy primitive
  val letSrcView = CGenLet(ParamDef(IntType(32)), IdxS(ParamUse(letp), Constant(1, IntType(32))), Id(ParamUse(letp)))

  @Test
  def letViewTypeTest(): Unit = {
    assert(letEager.rightmostViewType.isInstanceOf[EagerTypeT])
    assert(letSrcView.rightmostViewType.isInstanceOf[LazyTypeT])
  }

  // a function's view type is determined by the function body,
  //  which should be determined correctly
  @Test
  def lambdaViewTypeTest(): Unit = {
    // case 0: function body is a eager primitive as
    // Id has a src view type
    val p = ParamDef(IntType(32))
    val lam0 = CLambda(p, Id(ParamUse(p)))
    assert(lam0.t.viewType.isInstanceOf[EagerTypeT])

    // case 1: function body is a let expression and returns a call to a eager primitive
    val lam1 = CLambda(p, letEager)

    assert(lam1.t.viewType.isInstanceOf[EagerTypeT])

    // case 2: function body is a let expression and returns a call to a lazy primitive
    // ArrayAccess has a src view type
    val lam2 = CLambda(p, letSrcView)

    assert(lam2.t.viewType.isInstanceOf[LazyTypeT])
  }

  // map's view type is up to its input function
  @Test
  def mapViewTypeTest(): Unit = {
    val p = ParamDef(IntType(32))
    val parr = ParamDef(ArrayType(IntType(32), 8))
    val f0 = CLambda(p, sqrt(ParamUse(p)))
    val map0 = Map(f0, ParamUse(parr))
    val lam0 = CLambda(parr, map0)
    assert(lam0.t.viewType.isInstanceOf[EagerTypeT])

    val map1 = Map(CLambda(letp, letEager), ParamUse(parr))
    val lam1 = CLambda(parr, map1)
    assert(lam1.t.viewType.isInstanceOf[EagerTypeT])

    val map2 = Map(f0, ParamUse(parr))
    val lam2 = CLambda(parr, map2)
    assert(lam2.t.viewType.isInstanceOf[DestViewTypeT])

    val map3 = Map(CLambda(letp, letEager), ParamUse(parr))
    val lam3 = CLambda(parr, map3)
    assert(lam3.t.viewType.isInstanceOf[DestViewTypeT])

    val map4 = Map(CLambda(letp, letSrcView), ParamUse(parr))
    val lam4 = CLambda(parr, map2)
    assert(lam4.t.viewType.isInstanceOf[LazyTypeT])
  }
}
