package core

import cGen.{ArrayType, DoubleType}
import org.junit.Test

class TypeCheckerTest {

  @Test
  def testLambda(): Unit = {
    val inputType = BuiltinDataType()
    val tree =
      {
        val p = ParamDef(UnknownType)
        Lambda(p, ParamUse(p))
      }.call(Value(inputType))
    val rebuiltTree = TypeChecker.check(tree)
    assert(!rebuiltTree.hasUnknownType)
    assert(rebuiltTree.t == inputType)
  }

  @Test
  def testTypeLambda(): Unit = {
    val inputType = BuiltinDataType()
    val tree =
      {
        val dtv = BuiltinDataTypeVar()
        TypeLambda(
          dtv,
          {
            val p = ParamDef(dtv)
            Lambda(p, ParamUse(p))
          }
        )
      }.call(Value(BuiltinDataType()))
    val rebuiltTree = TypeChecker.check(tree)
    assert(!rebuiltTree.hasUnknownType)
    assert(rebuiltTree.t == inputType)
  }

  @Test
  def testUnresolvableArithmetic(): Unit = {
    val t = ArithTypeVar()
    val constraints = Seq(
      (t, ArithType(27)),
      (t, ArithType(1044484))
    )
    assert(TypeChecker.solveConstraints(constraints).isEmpty)
  }
}
