library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.common.all;

entity tuple is
    port(
        clk: in std_logic;
        reset: in std_logic;
        port_in_1_data: in data_word_type;
        port_in_1_last: in last_1_type;
        port_in_1_valid: in std_logic;
        port_in_1_ready: out ready_1_type;
        port_in_2_data: in data_word_type;
        port_in_2_last: in last_1_type;
        port_in_2_valid: in std_logic;
        port_in_2_ready: out ready_1_type;
        port_out_data: out data_word_tuple_type;
        port_out_last: out last_1_type;
        port_out_valid: out std_logic;
        port_out_ready: in ready_1_type
    );
end tuple;

architecture behavioral of tuple is
begin

    -- TODO special case if input 1 and input 2 have different length!

    port_out_data.t0 <= port_in_1_data;
    port_out_data.t1 <= port_in_2_data;

    last_signal: process(port_in_1_last, port_in_2_last)
        variable out_last: std_logic_vector(port_out_last'range) := (others => '1');
    begin
        out_last := (others => '1');
        out_last(out_last'high downto out_last'high - port_in_1_last'length + 1) := port_in_1_last and out_last(out_last'high downto out_last'high - port_in_1_last'length + 1);
        out_last(out_last'high downto out_last'high - port_in_2_last'length + 1) := port_in_2_last and out_last(out_last'high downto out_last'high - port_in_2_last'length + 1);
        port_out_last <= out_last;
    end process;

    port_out_valid <= port_in_1_valid and port_in_2_valid;

    port_in_1_ready <= port_out_ready(port_out_ready'high downto port_out_ready'high - port_in_1_ready'length + 1) when port_in_2_valid = '1' else (others => '0');

    port_in_2_ready <= port_out_ready(port_out_ready'high downto port_out_ready'high - port_in_2_ready'length + 1) when port_in_1_valid = '1' else (others => '0');

end behavioral;
